# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at http://www.comet.ml
#  Copyright (C) 2021 Comet ML INC
#  This file can not be copied and/or distributed without the express
#  permission of Comet ML Inc.
# *******************************************************
EVENT_PREDICTION_ID = "predictionId"
EVENT_TIMESTAMP = "timestamp"
EVENT_FEATURES = "features"
EVENT_PREDICTION = "prediction"
EVENT_PREDICTION_VALUE = "value"
EVENT_PREDICTION_PROBABILITY = "probability"
EVENT_WORKSPACE_NAME = "workspaceName"
EVENT_MODEL_NAME = "modelName"
EVENT_MODEL_VERSION = "modelVersion"
